# -*- Mode: Python; py-indent-offset: 4 -*-
# vim: tabstop=4 shiftwidth=4 expandtab
#
# Copyright (C) 2025 James Henstridge <james@jamesh.id.au>
#
#   _enum.py: base classes for enumeration and flags types.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import enum
import warnings

from . import _gi


class GEnumMeta(enum.EnumMeta):
    def __new__(metacls, name, bases, classdict, /, **kwargs):
        enum_class = super().__new__(metacls, name, bases, classdict, **kwargs)

        # If __gtype__ is not set, this is a new enum or flags defined
        # from Python. Register a new GType for it.
        if "__gtype__" not in enum_class.__dict__:
            type_name = enum_class.__dict__.get("__gtype_name__")
            _gi.enum_register(enum_class, type_name)

        return enum_class

    @property
    def __enum_values__(self):
        from gi import PyGIDeprecationWarning

        warnings.warn(
            'Calling "__enum_values__" on enums is deprecated.', PyGIDeprecationWarning
        )
        return {m.value: m for m in self}


class GEnum(enum.IntEnum, metaclass=GEnumMeta):
    __module__ = _gi.__name__
    __gtype__ = None


class GFlagsMeta(enum.EnumMeta):
    def __new__(metacls, name, bases, classdict, /, **kwargs):
        flags_class = super().__new__(metacls, name, bases, classdict, **kwargs)

        # If __gtype__ is not set, this is a new enum or flags defined
        # from Python. Register a new GType for it.
        if "__gtype__" not in flags_class.__dict__:
            type_name = flags_class.__dict__.get("__gtype_name__")
            _gi.flags_register(flags_class, type_name)

        return flags_class

    @property
    def __flags_values__(self):
        from gi import PyGIDeprecationWarning

        warnings.warn(
            'Calling "__flags_values__" on flags is deprecated. ',
            PyGIDeprecationWarning,
        )
        return {m.value: m for m in self}


class GFlags(enum.IntFlag, metaclass=GFlagsMeta):
    __module__ = _gi.__name__
    __gtype__ = None



# Copyright (C) 2014 Simon Feltman <sfeltman@gnome.org>
#
#   _error.py: GError Python implementation
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA


# NOTE: This file should not have any dependencies on introspection libs
# like gi.repository.GLib because it would cause a circular dependency.
# Developers wanting to use the GError class in their applications should
# use gi.repository.GLib.GError

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing_extensions import Self


class GError(RuntimeError):
    message: str
    domain: str
    code: int

    def __init__(
        self, message: str = "unknown error", domain: str = "pygi-error", code: int = 0
    ) -> None:
        super().__init__(message)
        self.message = message
        self.domain = domain
        self.code = code

    def __str__(self) -> str:
        return f"{self.domain:s}: {self.message:s} ({self.code:d})"

    def __repr__(self) -> str:
        return f"{GError.__module__.rsplit('.', 1)[-1]:s}.{GError.__name__:s}('{self.message:s}', '{self.domain:s}', {self.code:d})"

    def copy(self) -> Self:
        return GError(self.message, self.domain, self.code)

    def matches(self, domain: str | int, code: int) -> bool:
        """Placeholder that will be monkey patched in GLib overrides."""
        raise NotImplementedError

    @staticmethod
    def new_literal(domain: int, message: str, code: int) -> Self:
        """Placeholder that will be monkey patched in GLib overrides."""
        raise NotImplementedError



# Copyright (C) 2010 Ignacio Casal Quinteiro <icq@gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import asyncio
import warnings

from .._ossighelper import register_sigint_fallback, get_event_loop
from ..overrides import (
    override,
    deprecated_attr,
    deprecated_init,
    wrap_list_store_equal_func,
    wrap_list_store_sort_func,
)
from ..module import (
    get_introspection_module,
    get_fallback_platform_specific_module_symbol,
    get_platform_specific_module,
)
from gi import PyGIWarning


from gi.repository import GLib
from gi.repository import GObject

from typing import Generic, TypeVar, overload
from collections.abc import Callable, Generator, Sequence

import sys

Gio = get_introspection_module("Gio")

__all__ = []


class ActionMap(Gio.ActionMap):
    def add_action_entries(self, entries, user_data=None):
        """The ``add_action_entries()`` method is a convenience function for creating
        multiple :class:`~gi.repository.Gio.SimpleAction` instances and adding them
        to a :class:`~gi.repository.Gio.ActionMap`.
        Each action is constructed as per one entry.

        :param list entries:
            List of entry tuples for :meth:`add_action` method. The entry tuple can
            vary in size with the following information:

            * The name of the action. Must be specified.
            * The callback to connect to the "activate" signal of the
              action. Since GLib 2.40, this can be ``None`` for stateful
              actions, in which case the default handler is used. For
              boolean-stated actions with no parameter, this is a toggle.
              For other state types (and parameter type equal to the state
              type) this will be a function that just calls change_state
              (which you should provide).
            * The type of the parameter that must be passed to the activate
              function for this action, given as a single :class:`~gi.repository.GLib.Variant` type
              string (or ``None`` for no parameter)
            * The initial state for this action, given in GLib.Variant text
              format. The state is parsed with no extra type information, so
              type tags must be added to the string if they are necessary.
              Stateless actions should give ``None`` here.
            * The callback to connect to the "change-state" signal of the
              action. All stateful actions should provide a handler here;
              stateless actions should not.

        :param user_data:
            The user data for signal connections, or ``None``
        """
        try:
            iter(entries)
        except TypeError:
            raise TypeError("entries must be iterable")

        def _process_action(
            name, activate=None, parameter_type=None, state=None, change_state=None
        ):
            if parameter_type:
                if not GLib.VariantType.string_is_valid(parameter_type):
                    msg = f"The type string '{parameter_type}' given as the parameter type for action '{name}' is not a valid GVariant type string. "
                    raise TypeError(msg)
                variant_parameter = GLib.VariantType.new(parameter_type)
            else:
                variant_parameter = None

            if state is not None:
                # stateful action
                variant_state = GLib.Variant.parse(None, state, None, None)
                action = Gio.SimpleAction.new_stateful(
                    name, variant_parameter, variant_state
                )
                if change_state is not None:
                    action.connect("change-state", change_state, user_data)
            else:
                # stateless action
                if change_state is not None:
                    msg = f"Stateless action '{name}' should give None for 'change_state', not '{change_state}'."
                    raise ValueError(msg)
                action = Gio.SimpleAction(name=name, parameter_type=variant_parameter)

            if activate is not None:
                action.connect("activate", activate, user_data)
            self.add_action(action)

        for entry in entries:
            # using inner function above since entries can leave out optional arguments
            _process_action(*entry)


ActionMap = override(ActionMap)
__all__.append("ActionMap")


class Application(Gio.Application):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._asyncio_tasks = set()

    def run(self, *args, **kwargs):
        with (
            register_sigint_fallback(self.quit),
            get_event_loop(GLib.MainContext.default()).running(self.quit),
        ):
            return Gio.Application.run(self, *args, **kwargs)

    def create_asyncio_task(self, coro):
        """Safely create an asyncio task. The application will not quit until the
        task completes. For potentially longer running tasks, you should add
        cancellation logic to abort a task when it is not needed anymore (e.g.
        cancelling it from the Gtk.Window.do_unmap event).

        Note that python will only log a raised exception if the Task is
        destroyed without the result having been collected. However, this does
        also not happen when the task is cancelled. As such, be careful to not
        cancel tasks that are already finished.

        You can deal with this by either only storing a weak reference to the
        Task, by explicitly collecting the result, or by only cancelling it if
        it is not done already.
        """
        # First create the task (in case it raises an exception)
        task = asyncio.create_task(coro)

        self.hold()

        # Store a reference to the task so that it cannot be garbage collected
        self._asyncio_tasks.add(task)

        def done_cb(task):
            nonlocal self
            self._asyncio_tasks.discard(task)
            self.release()

        task.add_done_callback(done_cb)
        return task


Application = override(Application)
__all__.append("Application")


def _warn_init(cls, instead=None):
    def new_init(self, *args, **kwargs):
        super(cls, self).__init__(*args, **kwargs)
        name = cls.__module__.rsplit(".", 1)[-1] + "." + cls.__name__
        if instead:
            warnings.warn(
                (f"{name} shouldn't be instantiated directly, use {instead} instead."),
                PyGIWarning,
                stacklevel=2,
            )
        else:
            warnings.warn(
                f"{name} shouldn't be instantiated directly.",
                PyGIWarning,
                stacklevel=2,
            )

    return new_init


@override
class VolumeMonitor(Gio.VolumeMonitor):
    # https://bugzilla.gnome.org/show_bug.cgi?id=744690
    __init__ = _warn_init(Gio.VolumeMonitor, "Gio.VolumeMonitor.get()")


__all__.append("VolumeMonitor")


@override
class DBusAnnotationInfo(Gio.DBusAnnotationInfo):
    __init__ = _warn_init(Gio.DBusAnnotationInfo)


__all__.append("DBusAnnotationInfo")


@override
class DBusArgInfo(Gio.DBusArgInfo):
    __init__ = _warn_init(Gio.DBusArgInfo)


__all__.append("DBusArgInfo")


@override
class DBusMethodInfo(Gio.DBusMethodInfo):
    __init__ = _warn_init(Gio.DBusMethodInfo)


__all__.append("DBusMethodInfo")


@override
class DBusSignalInfo(Gio.DBusSignalInfo):
    __init__ = _warn_init(Gio.DBusSignalInfo)


__all__.append("DBusSignalInfo")


@override
class DBusInterfaceInfo(Gio.DBusInterfaceInfo):
    __init__ = _warn_init(Gio.DBusInterfaceInfo)


__all__.append("DBusInterfaceInfo")


@override
class DBusNodeInfo(Gio.DBusNodeInfo):
    __init__ = _warn_init(Gio.DBusNodeInfo)


__all__.append("DBusNodeInfo")


class FileEnumerator(Gio.FileEnumerator):
    def __iter__(self):
        return self

    def __next__(self):
        file_info = self.next_file(None)

        if file_info is not None:
            return file_info
        raise StopIteration


FileEnumerator = override(FileEnumerator)
__all__.append("FileEnumerator")


class MenuItem(Gio.MenuItem):
    def set_attribute(self, attributes):
        for name, format_string, value in attributes:
            self.set_attribute_value(name, GLib.Variant(format_string, value))


MenuItem = override(MenuItem)
__all__.append("MenuItem")


class Settings(Gio.Settings):
    """Provide dictionary-like access to GLib.Settings."""

    __init__ = deprecated_init(
        Gio.Settings.__init__, arg_names=("schema", "path", "backend")
    )

    def __contains__(self, key):
        return key in self.list_keys()

    def __len__(self):
        return len(self.list_keys())

    def __iter__(self):
        yield from self.list_keys()

    def __bool__(self):
        # for "if mysettings" we don't want a dictionary-like test here, just
        # if the object isn't None
        return True

    def __getitem__(self, key):
        # get_value() aborts the program on an unknown key
        if key not in self:
            raise KeyError(f"unknown key: {key!r}")

        return self.get_value(key).unpack()

    def __setitem__(self, key, value):
        # set_value() aborts the program on an unknown key
        if key not in self:
            raise KeyError(f"unknown key: {key!r}")

        # determine type string of this key
        range = self.get_range(key)
        type_ = range.get_child_value(0).get_string()
        v = range.get_child_value(1)
        if type_ == "type":
            # v is boxed empty array, type of its elements is the allowed value type
            type_str = v.get_child_value(0).get_type_string()
            assert type_str.startswith("a")
            type_str = type_str[1:]
        elif type_ == "enum":
            # v is an array with the allowed values
            assert v.get_child_value(0).get_type_string().startswith("a")
            type_str = v.get_child_value(0).get_child_value(0).get_type_string()
            allowed = v.unpack()
            if value not in allowed:
                raise ValueError(f"value {value} is not an allowed enum ({allowed})")
        elif type_ == "range":
            tuple_ = v.get_child_value(0)
            type_str = tuple_.get_child_value(0).get_type_string()
            min_, max_ = tuple_.unpack()
            if value < min_ or value > max_:
                raise ValueError(f"value {value} not in range ({min_} - {max_})")
        else:
            raise NotImplementedError(
                "Cannot handle allowed type range class " + str(type_)
            )

        self.set_value(key, GLib.Variant(type_str, value))

    def keys(self):
        return self.list_keys()


Settings = override(Settings)
__all__.append("Settings")


class _DBusProxyMethodCall:
    """Helper class to implement DBusProxy method calls."""

    def __init__(self, dbus_proxy, method_name):
        self.dbus_proxy = dbus_proxy
        self.method_name = method_name

    def __async_result_handler(self, obj, result, user_data):
        (result_callback, error_callback, real_user_data) = user_data
        try:
            ret = obj.call_finish(result)
        except Exception:
            _etype, e = sys.exc_info()[:2]
            # return exception as value
            if error_callback:
                error_callback(obj, e, real_user_data)
            else:
                result_callback(obj, e, real_user_data)
            return

        result_callback(obj, self._unpack_result(ret), real_user_data)

    def __call__(
        self,
        *args,
        result_handler=None,
        error_handler=None,
        user_data=None,
        flags=0,
        timeout=-1,
    ):
        # the first positional argument is the signature, unless we are calling
        # a method without arguments; then signature is implied to be '()'.
        if args:
            signature = args[0]
            args = args[1:]
            if not isinstance(signature, str):
                raise TypeError(
                    f"first argument must be the method signature string: {signature!r}"
                )
        else:
            signature = "()"

        arg_variant = GLib.Variant(signature, tuple(args))

        if result_handler is not None:
            # asynchronous call
            user_data = (result_handler, error_handler, user_data)
            self.dbus_proxy.call(
                self.method_name,
                arg_variant,
                flags,
                timeout,
                None,
                self.__async_result_handler,
                user_data,
            )
        else:
            # synchronous call
            result = self.dbus_proxy.call_sync(
                self.method_name, arg_variant, flags, timeout, None
            )
            return self._unpack_result(result)
        return None

    @classmethod
    def _unpack_result(klass, result):
        """Convert a D-BUS return variant into an appropriate return value."""
        result = result.unpack()

        # to be compatible with standard Python behaviour, unbox
        # single-element tuples and return None for empty result tuples
        if len(result) == 1:
            result = result[0]
        elif len(result) == 0:
            result = None

        return result


class DBusProxy(Gio.DBusProxy):
    """Provide comfortable and pythonic method calls.

    This marshalls the method arguments into a GVariant, invokes the
    call_sync() method on the DBusProxy object, and unmarshalls the result
    GVariant back into a Python tuple.

    The first argument always needs to be the D-Bus signature tuple of the
    method call. Example:

      proxy = Gio.DBusProxy.new_sync(...)
      result = proxy.MyMethod('(is)', 42, 'hello')

    The exception are methods which take no arguments, like
    proxy.MyMethod('()'). For these you can omit the signature and just write
    proxy.MyMethod().

    Optional keyword arguments:

    - timeout: timeout for the call in milliseconds (default to D-Bus timeout)

    - flags: Combination of Gio.DBusCallFlags.*

    - result_handler: Do an asynchronous method call and invoke
         result_handler(proxy_object, result, user_data) when it finishes.

    - error_handler: If the asynchronous call raises an exception,
      error_handler(proxy_object, exception, user_data) is called when it
      finishes. If error_handler is not given, result_handler is called with
      the exception object as result instead.

    - user_data: Optional user data to pass to result_handler for
      asynchronous calls.

    Example for asynchronous calls:

      def mymethod_done(proxy, result, user_data):
          if isinstance(result, Exception):
              # handle error
          else:
              # do something with result

      proxy.MyMethod('(is)', 42, 'hello',
          result_handler=mymethod_done, user_data='data')
    """

    def __getattr__(self, name):
        if name.startswith("do_"):
            return self[name]
        return _DBusProxyMethodCall(self, name)


DBusProxy = override(DBusProxy)
__all__.append("DBusProxy")


ObjectItemType = TypeVar("ObjectItemType", bound=GObject.Object)


class ListModel(Gio.ListModel, Generic[ObjectItemType]):
    @overload
    def __getitem__(self, key: slice) -> list[ObjectItemType]: ...

    @overload
    def __getitem__(self, key: int) -> ObjectItemType: ...

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self.get_item(i) for i in range(*key.indices(len(self)))]
        if isinstance(key, int):
            if key < 0:
                key += len(self)
            if key < 0:
                raise IndexError
            ret = self.get_item(key)
            if ret is None:
                raise IndexError
            return ret
        raise TypeError

    def __contains__(self, item: ObjectItemType) -> bool:
        pytype = self.get_item_type().pytype
        if not isinstance(item, pytype):
            raise TypeError(f"Expected type {pytype.__module__}.{pytype.__name__}")
        return any(i == item for i in self)

    def __len__(self) -> int:
        return self.get_n_items()

    def __iter__(self) -> Generator[ObjectItemType, None, None]:
        for i in range(len(self)):
            yield self.get_item(i)


ListModel = override(ListModel)
__all__.append("ListModel")


class ListStore(Gio.ListStore, Generic[ObjectItemType]):
    # Describing the variadic arguments requires TypeVarTuple and unpacking syntax in type annotation:
    #   compare_func: Callable[Concatenate[ObjectItemType, ObjectItemType, *TypedVarTuple('Args')], int]
    #   *user_data: *TypedVarTuple('Args')
    # Since those are only available on Python ≥ 3.11, let's keep the arguments untyped for now.
    def sort(self, compare_func: Callable[..., int], *user_data) -> None:
        compare_func = wrap_list_store_sort_func(compare_func)
        return super().sort(compare_func, *user_data)

    def insert_sorted(
        self, item: ObjectItemType, compare_func: Callable[..., int], *user_data
    ) -> None:
        compare_func = wrap_list_store_sort_func(compare_func)
        return super().insert_sorted(item, compare_func, *user_data)

    def __delitem__(self, key: int | slice) -> None:
        if isinstance(key, slice):
            start, stop, step = key.indices(len(self))
            if step == 1:
                self.splice(start, max(stop - start, 0), [])
            elif step == -1:
                self.splice(stop + 1, max(start - stop, 0), [])
            else:
                for i in sorted(range(start, stop, step), reverse=True):
                    self.remove(i)
        elif isinstance(key, int):
            if key < 0:
                key += len(self)
            if key < 0 or key >= len(self):
                raise IndexError
            self.remove(key)
        else:
            raise TypeError

    @overload
    def __setitem__(self, key: slice, value: Sequence[ObjectItemType]) -> None: ...

    @overload
    def __setitem__(self, key: int, value: ObjectItemType) -> None: ...

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            pytype = self.get_item_type().pytype
            valuelist = []
            for v in value:
                if not isinstance(v, pytype):
                    raise TypeError(
                        f"Expected type {pytype.__module__}.{pytype.__name__}"
                    )
                valuelist.append(v)

            start, stop, step = key.indices(len(self))
            if step == 1:
                self.splice(start, max(stop - start, 0), valuelist)
            else:
                indices = list(range(start, stop, step))
                if len(indices) != len(valuelist):
                    raise ValueError

                if step == -1:
                    self.splice(stop + 1, max(start - stop, 0), valuelist[::-1])
                else:
                    for i, v in zip(indices, valuelist):
                        self.splice(i, 1, [v])
        elif isinstance(key, int):
            if key < 0:
                key += len(self)
            if key < 0 or key >= len(self):
                raise IndexError

            pytype = self.get_item_type().pytype
            if not isinstance(value, pytype):
                raise TypeError(f"Expected type {pytype.__module__}.{pytype.__name__}")

            self.splice(key, 1, [value])
        else:
            raise TypeError

    def find_with_equal_func(self, item, equal_func, *user_data):
        # find_with_equal_func() is not suited for language bindings,
        # see: https://gitlab.gnome.org/GNOME/glib/-/issues/2447.
        # Use find_with_equal_func_full() instead.
        equal_func = wrap_list_store_equal_func(equal_func)
        return self.find_with_equal_func_full(item, equal_func, *user_data)


ListStore = override(ListStore)
__all__.append("ListStore")


class DataInputStream(Gio.DataInputStream):
    def __iter__(self):
        return self

    def __next__(self):
        line = self.read_line_utf8(None)[0]

        if line is not None:
            return line
        raise StopIteration


DataInputStream = override(DataInputStream)
__all__.append("DataInputStream")


class File(Gio.File):
    def __fspath__(self):
        path = self.peek_path()
        # A file isn't guaranteed to have a path associated and returning
        # `None` here will result in a `TypeError` trying to subscribe to it.
        if path is None or path == "":
            raise TypeError("File has no associated path.")
        return path


File = override(File)
__all__.append("File")


GioPlatform = get_platform_specific_module(Gio)

if GioPlatform:
    # Add support for using platform-specific Gio symbols.
    gio_globals = globals()

    for attr in dir(GioPlatform):
        if attr.startswith("_"):
            continue

        original_attr, wrapper_attr = get_fallback_platform_specific_module_symbol(
            Gio, GioPlatform, attr
        )

        if (
            wrapper_attr == attr
            and (GLib.MAJOR_VERSION, GLib.MINOR_VERSION) >= (2, 86)
            and hasattr(Gio, wrapper_attr)
        ):
            warnings.warn(
                (
                    f"Name conflict for platform-specific symbol {GioPlatform._namespace}.{attr}."
                    + " No wrapper will be created."
                ),
                PyGIWarning,
                stacklevel=2,
            )

        if (
            wrapper_attr in __all__
            or wrapper_attr in gio_globals
            or hasattr(Gio, wrapper_attr)
        ):
            continue

        gio_globals[wrapper_attr] = original_attr
        deprecated_attr(
            Gio._namespace, wrapper_attr, f"{GioPlatform._namespace}.{attr}"
        )
        __all__.append(wrapper_attr)



# -*- Mode: Python; py-indent-offset: 4 -*-
# vim: tabstop=4 shiftwidth=4 expandtab
#
# Copyright (C) 2012 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
# Copyright (C) 2012-2013 Simon Feltman <sfeltman@src.gnome.org>
# Copyright (C) 2012 Bastian Winkler <buz@netbuz.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

from __future__ import annotations

import functools
import warnings
import typing
from collections import namedtuple

from gi.overrides import override, deprecated_attr
from gi.repository import GLib
from gi import PyGIDeprecationWarning
from gi import _propertyhelper as propertyhelper
from gi import _signalhelper as signalhelper
from gi import _gi

# Typing relies on https://github.com/pygobject/pygobject-stubs.
if typing.TYPE_CHECKING:
    # Import stubs for type checking this file.
    # FIXME: For IDE typing support, you need to copy
    # pygobject-stubs/src/gi-stubs/repository/*.pyi to gi/repository/
    from gi.repository import GObject as GObjectModule

    # Type annotations cannot have `GObject.` prefix because they are copied into
    # GObject stubs module which cannot refer to itself. Use type aliases.
    GType: typing.TypeAlias = GObjectModule.GType
    Object = GObjectModule.Object
    ConnectFlags = GObjectModule.ConnectFlags
else:
    from gi.module import get_introspection_module

    GObjectModule = get_introspection_module("GObject")


__all__ = []


# API aliases for backwards compatibility
for name in [
    "markup_escape_text",
    "get_application_name",
    "set_application_name",
    "get_prgname",
    "set_prgname",
    "main_depth",
    "filename_display_basename",
    "filename_display_name",
    "filename_from_utf8",
    "uri_list_extract_uris",
    "MainLoop",
    "MainContext",
    "main_context_default",
    "source_remove",
    "Source",
    "Idle",
    "Timeout",
    "PollFD",
    "idle_add",
    "timeout_add",
    "timeout_add_seconds",
    "io_add_watch",
    "child_watch_add",
    "get_current_time",
    "spawn_async",
]:
    globals()[name] = getattr(GLib, name)
    deprecated_attr("GObject", name, "GLib." + name)
    __all__.append(name)

# deprecated constants
for name in [
    "PRIORITY_DEFAULT",
    "PRIORITY_DEFAULT_IDLE",
    "PRIORITY_HIGH",
    "PRIORITY_HIGH_IDLE",
    "PRIORITY_LOW",
    "IO_IN",
    "IO_OUT",
    "IO_PRI",
    "IO_ERR",
    "IO_HUP",
    "IO_NVAL",
    "IO_STATUS_ERROR",
    "IO_STATUS_NORMAL",
    "IO_STATUS_EOF",
    "IO_STATUS_AGAIN",
    "IO_FLAG_APPEND",
    "IO_FLAG_NONBLOCK",
    "IO_FLAG_IS_READABLE",
    "IO_FLAG_IS_WRITEABLE",
    "IO_FLAG_IS_SEEKABLE",
    "IO_FLAG_MASK",
    "IO_FLAG_GET_MASK",
    "IO_FLAG_SET_MASK",
    "SPAWN_LEAVE_DESCRIPTORS_OPEN",
    "SPAWN_DO_NOT_REAP_CHILD",
    "SPAWN_SEARCH_PATH",
    "SPAWN_STDOUT_TO_DEV_NULL",
    "SPAWN_STDERR_TO_DEV_NULL",
    "SPAWN_CHILD_INHERITS_STDIN",
    "SPAWN_FILE_AND_ARGV_ZERO",
    "OPTION_FLAG_HIDDEN",
    "OPTION_FLAG_IN_MAIN",
    "OPTION_FLAG_REVERSE",
    "OPTION_FLAG_NO_ARG",
    "OPTION_FLAG_FILENAME",
    "OPTION_FLAG_OPTIONAL_ARG",
    "OPTION_FLAG_NOALIAS",
    "OPTION_ERROR_UNKNOWN_OPTION",
    "OPTION_ERROR_BAD_VALUE",
    "OPTION_ERROR_FAILED",
    "OPTION_REMAINING",
    "glib_version",
]:
    with warnings.catch_warnings():
        # TODO: this uses deprecated Glib attributes, silence for now
        warnings.simplefilter("ignore", PyGIDeprecationWarning)
        globals()[name] = getattr(GLib, name)
    deprecated_attr("GObject", name, "GLib." + name)
    __all__.append(name)


for name in [
    "G_MININT8",
    "G_MAXINT8",
    "G_MAXUINT8",
    "G_MININT16",
    "G_MAXINT16",
    "G_MAXUINT16",
    "G_MININT32",
    "G_MAXINT32",
    "G_MAXUINT32",
    "G_MININT64",
    "G_MAXINT64",
    "G_MAXUINT64",
]:
    new_name = name.split("_", 1)[-1]
    globals()[name] = getattr(GLib, new_name)
    deprecated_attr("GObject", name, "GLib." + new_name)
    __all__.append(name)

# these are not currently exported in GLib gir, presumably because they are
# platform dependent; so get them from our static bindings
for name in [
    "G_MINFLOAT",
    "G_MAXFLOAT",
    "G_MINDOUBLE",
    "G_MAXDOUBLE",
    "G_MINSHORT",
    "G_MAXSHORT",
    "G_MAXUSHORT",
    "G_MININT",
    "G_MAXINT",
    "G_MAXUINT",
    "G_MINLONG",
    "G_MAXLONG",
    "G_MAXULONG",
    "G_MAXSIZE",
    "G_MINSSIZE",
    "G_MAXSSIZE",
    "G_MINOFFSET",
    "G_MAXOFFSET",
]:
    new_name = name.split("_", 1)[-1]
    globals()[name] = getattr(GLib, new_name)
    deprecated_attr("GObject", name, "GLib." + new_name)
    __all__.append(name)


TYPE_INVALID: GType = GObjectModule.type_from_name("invalid")
TYPE_NONE: GType = GObjectModule.type_from_name("void")
TYPE_INTERFACE: GType = GObjectModule.type_from_name("GInterface")
TYPE_CHAR: GType = GObjectModule.type_from_name("gchar")
TYPE_UCHAR: GType = GObjectModule.type_from_name("guchar")
TYPE_BOOLEAN: GType = GObjectModule.type_from_name("gboolean")
TYPE_INT: GType = GObjectModule.type_from_name("gint")
TYPE_UINT: GType = GObjectModule.type_from_name("guint")
TYPE_LONG: GType = GObjectModule.type_from_name("glong")
TYPE_ULONG: GType = GObjectModule.type_from_name("gulong")
TYPE_INT64: GType = GObjectModule.type_from_name("gint64")
TYPE_UINT64: GType = GObjectModule.type_from_name("guint64")
TYPE_ENUM: GType = GObjectModule.type_from_name("GEnum")
TYPE_FLAGS: GType = GObjectModule.type_from_name("GFlags")
TYPE_FLOAT: GType = GObjectModule.type_from_name("gfloat")
TYPE_DOUBLE: GType = GObjectModule.type_from_name("gdouble")
TYPE_STRING: GType = GObjectModule.type_from_name("gchararray")
TYPE_POINTER: GType = GObjectModule.type_from_name("gpointer")
TYPE_BOXED: GType = GObjectModule.type_from_name("GBoxed")
TYPE_PARAM: GType = GObjectModule.type_from_name("GParam")
TYPE_OBJECT: GType = GObjectModule.type_from_name("GObject")
TYPE_PYOBJECT: GType = GObjectModule.type_from_name("PyObject")
TYPE_GTYPE: GType = GObjectModule.type_from_name("GType")
TYPE_STRV: GType = GObjectModule.type_from_name("GStrv")
TYPE_VARIANT: GType = GObjectModule.type_from_name("GVariant")
TYPE_GSTRING: GType = GObjectModule.type_from_name("GString")
TYPE_VALUE: GType = GObjectModule.Value.__gtype__
TYPE_UNICHAR: GType = TYPE_UINT
__all__ += [
    "TYPE_BOOLEAN",
    "TYPE_BOXED",
    "TYPE_CHAR",
    "TYPE_DOUBLE",
    "TYPE_ENUM",
    "TYPE_FLAGS",
    "TYPE_FLOAT",
    "TYPE_GSTRING",
    "TYPE_GTYPE",
    "TYPE_INT",
    "TYPE_INT64",
    "TYPE_INTERFACE",
    "TYPE_INVALID",
    "TYPE_LONG",
    "TYPE_NONE",
    "TYPE_OBJECT",
    "TYPE_PARAM",
    "TYPE_POINTER",
    "TYPE_PYOBJECT",
    "TYPE_STRING",
    "TYPE_STRV",
    "TYPE_UCHAR",
    "TYPE_UINT",
    "TYPE_UINT64",
    "TYPE_ULONG",
    "TYPE_UNICHAR",
    "TYPE_VALUE",
    "TYPE_VARIANT",
]


# Deprecated, use GLib directly
for name in ["Pid", "GError", "OptionGroup", "OptionContext"]:
    globals()[name] = getattr(GLib, name)
    deprecated_attr("GObject", name, "GLib." + name)
    __all__.append(name)


# Deprecated, use: GObject.ParamFlags.* directly
for name in [
    "PARAM_CONSTRUCT",
    "PARAM_CONSTRUCT_ONLY",
    "PARAM_LAX_VALIDATION",
    "PARAM_READABLE",
    "PARAM_WRITABLE",
]:
    new_name = name.split("_", 1)[-1]
    globals()[name] = getattr(GObjectModule.ParamFlags, new_name)
    deprecated_attr("GObject", name, "GObject.ParamFlags." + new_name)
    __all__.append(name)

# PARAM_READWRITE should come from the gi module but cannot due to:
# https://gitlab.gnome.org/GNOME/gobject-introspection/issues/75
PARAM_READWRITE = GObjectModule.ParamFlags.READABLE | GObjectModule.ParamFlags.WRITABLE
deprecated_attr("GObject", "PARAM_READWRITE", "GObject.ParamFlags.READWRITE")
__all__.append("PARAM_READWRITE")


# Deprecated, use: GObject.SignalFlags.* directly
for name in [
    "SIGNAL_ACTION",
    "SIGNAL_DETAILED",
    "SIGNAL_NO_HOOKS",
    "SIGNAL_NO_RECURSE",
    "SIGNAL_RUN_CLEANUP",
    "SIGNAL_RUN_FIRST",
    "SIGNAL_RUN_LAST",
]:
    new_name = name.split("_", 1)[-1]
    globals()[name] = getattr(GObjectModule.SignalFlags, new_name)
    deprecated_attr("GObject", name, "GObject.SignalFlags." + new_name)
    __all__.append(name)

# Static types
GBoxed = _gi.GBoxed
GEnum = _gi.GEnum
GFlags = _gi.GFlags
GInterface = _gi.GInterface
GObject = _gi.GObject
GObjectWeakRef = _gi.GObjectWeakRef
GPointer = _gi.GPointer
GType = _gi.GType
Warning = _gi.Warning
__all__ += [
    "GBoxed",
    "GEnum",
    "GFlags",
    "GInterface",
    "GObject",
    "GObjectWeakRef",
    "GPointer",
    "GType",
    "Warning",
]


features = {"generic-c-marshaller": True}
list_properties = _gi.list_properties
new = _gi.new
pygobject_version = _gi.pygobject_version
threads_init = GLib.threads_init
type_register = _gi.type_register
__all__ += [
    "features",
    "list_properties",
    "new",
    "pygobject_version",
    "threads_init",
    "type_register",
]


class Value(GObjectModule.Value):
    def __init__(
        self, value_type: GType | None = None, py_value: typing.Any | None = None
    ) -> None:
        GObjectModule.Value.__init__(self)
        if value_type is not None:
            self.init(value_type)
            if py_value is not None:
                self.set_value(py_value)

    @property
    def __g_type(self):
        # XXX: This is the same as self.g_type, but the field marshalling
        # code is currently very slow.
        return _gi._gvalue_get_type(self)

    def set_boxed(self, boxed: GBoxed):
        if not self.__g_type.is_a(TYPE_BOXED):
            warnings.warn(
                "Calling set_boxed() on a non-boxed type deprecated",
                PyGIDeprecationWarning,
                stacklevel=2,
            )
        # Workaround the introspection marshalers inability to know
        # these methods should be marshaling boxed types. This is because
        # the type information is stored on the GValue.
        _gi._gvalue_set(self, boxed)

    def get_boxed(self) -> GBoxed:
        if not self.__g_type.is_a(TYPE_BOXED):
            warnings.warn(
                "Calling get_boxed() on a non-boxed type deprecated",
                PyGIDeprecationWarning,
                stacklevel=2,
            )
        return _gi._gvalue_get(self)

    def set_value(self, py_value: typing.Any) -> None:
        gtype = self.__g_type

        if gtype == TYPE_CHAR:
            self.set_char(py_value)
        elif gtype == TYPE_UCHAR:
            self.set_uchar(py_value)
        elif gtype == TYPE_STRING:
            if not isinstance(py_value, str) and py_value is not None:
                raise TypeError(f"Expected string but got {py_value}{type(py_value)}")
            _gi._gvalue_set(self, py_value)
        elif gtype == TYPE_PARAM:
            self.set_param(py_value)
        elif gtype.is_a(TYPE_FLAGS):
            self.set_flags(py_value)
        elif gtype == TYPE_POINTER:
            self.set_pointer(py_value)
        elif gtype == TYPE_GTYPE:
            self.set_gtype(py_value)
        elif gtype == TYPE_VARIANT:
            self.set_variant(py_value)
        else:
            # Fall back to _gvalue_set which handles some more cases
            # like fundamentals for which a converter is registered
            try:
                _gi._gvalue_set(self, py_value)
            except TypeError:
                if gtype == TYPE_INVALID:
                    raise TypeError("GObject.Value needs to be initialized first")
                raise

    def get_value(self) -> typing.Any:
        gtype = self.__g_type

        if gtype == TYPE_CHAR:
            return self.get_char()
        if gtype == TYPE_UCHAR:
            return self.get_uchar()
        if gtype == TYPE_PARAM:
            return self.get_param()
        if gtype.is_a(TYPE_ENUM):
            return self.get_enum()
        if gtype.is_a(TYPE_FLAGS):
            return self.get_flags()
        if gtype == TYPE_POINTER:
            return self.get_pointer()
        if gtype == TYPE_GTYPE:
            return self.get_gtype()
        if gtype == TYPE_VARIANT:
            # get_variant was missing annotations
            # https://gitlab.gnome.org/GNOME/glib/merge_requests/492
            return self.dup_variant()
        try:
            return _gi._gvalue_get(self)
        except TypeError:
            if gtype == TYPE_INVALID:
                return None
            raise

    def __repr__(self) -> str:
        return f"<Value ({self.__g_type.name}) {self.get_value()}>"


Value = override(Value)
__all__.append("Value")


def type_from_name(name: str) -> GType:
    type_ = GObjectModule.type_from_name(name)
    if type_ == TYPE_INVALID:
        raise RuntimeError(f"unknown type name: {name}")
    return type_


__all__.append("type_from_name")


def type_parent(type_: GType | Object) -> GType:
    parent = GObjectModule.type_parent(type_)
    if parent == TYPE_INVALID:
        raise RuntimeError("no parent for type")
    return parent


__all__.append("type_parent")


def _validate_type_for_signal_method(type_: GType | Object) -> None:
    if hasattr(type_, "__gtype__"):
        type_ = type_.__gtype__
    if not type_.is_instantiatable() and not type_.is_interface():
        raise TypeError(f"type must be instantiable or an interface, got {type_}")


def signal_list_ids(type_: GType | Object) -> tuple[int, ...]:
    _validate_type_for_signal_method(type_)
    return GObjectModule.signal_list_ids(type_)


__all__.append("signal_list_ids")


def signal_list_names(type_: GType | Object) -> tuple[str, ...]:
    ids = signal_list_ids(type_)
    return tuple(GObjectModule.signal_name(i) for i in ids)


__all__.append("signal_list_names")


def signal_lookup(name: str, type_: GType | Object) -> int:
    _validate_type_for_signal_method(type_)
    return GObjectModule.signal_lookup(name, type_)


__all__.append("signal_lookup")


SignalQuery = namedtuple(
    "SignalQuery",
    [
        "signal_id",
        "signal_name",
        "itype",
        "signal_flags",
        "return_type",
        # n_params',
        "param_types",
    ],
)


def signal_query(
    id_or_name: int | str, type_: GType | Object | None = None
) -> SignalQuery | None:
    if isinstance(id_or_name, str):
        assert type_ is not None, "type_ is required when looking up signal by name"
        signal_id = signal_lookup(id_or_name, type_)
    else:
        signal_id = id_or_name

    res = GObjectModule.signal_query(signal_id)
    assert res is not None

    if res.signal_id == 0:
        return None

    # Return a named tuple to allows indexing which is compatible with the
    # static bindings along with field like access of the gi struct.
    # Note however that the n_params was not returned from the static bindings
    # so we must skip over it.
    return SignalQuery(
        res.signal_id,
        res.signal_name,
        res.itype,
        res.signal_flags,
        res.return_type,
        tuple(res.param_types),
    )


__all__.append("signal_query")


class _HandlerBlockManager:
    def __init__(self, obj: Object, handler_id: int):
        self.obj = obj
        self.handler_id = handler_id

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        GObjectModule.signal_handler_unblock(self.obj, self.handler_id)


def signal_handler_block(obj: Object, handler_id: int) -> typing.ContextManager[None]:
    """Blocks the signal handler from being invoked until
    handler_unblock() is called.

    :param GObject.Object obj:
        Object instance to block handlers for.
    :param int handler_id:
        Id of signal to block.
    :returns:
        A context manager which optionally can be used to
        automatically unblock the handler:

    .. code-block:: python

        with GObject.signal_handler_block(obj, id):
            pass
    """
    GObjectModule.signal_handler_block(obj, handler_id)
    return _HandlerBlockManager(obj, handler_id)


__all__.append("signal_handler_block")


def signal_parse_name(
    detailed_signal: str, itype: GType | Object, force_detail_quark: bool
) -> tuple[int, int]:
    """Parse a detailed signal name into (signal_id, detail).

    :param str detailed_signal:
        Signal name which can include detail.
        For example: "notify:prop_name"
    :returns:
        Tuple of (signal_id, detail)
    :raises ValueError:
        If the given signal is unknown.
    """
    res, signal_id, detail = GObjectModule.signal_parse_name(
        detailed_signal, itype, force_detail_quark
    )
    if res:
        return signal_id, detail
    raise ValueError(f"{itype}: unknown signal name: {detailed_signal}")


__all__.append("signal_parse_name")


def remove_emission_hook(obj: Object, detailed_signal: str, hook_id: int) -> None:
    signal_id, _detail = signal_parse_name(detailed_signal, obj, True)
    GObjectModule.signal_remove_emission_hook(signal_id, hook_id)


__all__.append("remove_emission_hook")


# GObject accumulators with pure Python implementations
# These return a tuple of (continue_emission, accumulation_result)


def signal_accumulator_first_wins(ihint, return_accu, handler_return, user_data=None):
    # Stop emission but return the result of the last handler
    return (False, handler_return)


__all__.append("signal_accumulator_first_wins")


def signal_accumulator_true_handled(ihint, return_accu, handler_return, user_data=None):
    # Stop emission if the last handler returns True
    return (not handler_return, handler_return)


__all__.append("signal_accumulator_true_handled")


# Statically bound signal functions which need to clobber GI (for now)

add_emission_hook = _gi.add_emission_hook
signal_new = _gi.signal_new

__all__ += ["add_emission_hook", "signal_new"]


class _FreezeNotifyManager:
    def __init__(self, obj: Object):
        self.obj = obj

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.obj.thaw_notify()


def _signalmethod(func):
    # Function wrapper for signal functions used as instance methods.
    # This is needed when the signal functions come directly from GI.
    # (they are not already wrapped)
    @functools.wraps(func)
    def meth(*args, **kwargs):
        return func(*args, **kwargs)

    return meth


class Object(GObjectModule.Object):
    def _unsupported_method(self, *args, **kargs):
        raise RuntimeError("This method is currently unsupported.")

    def _unsupported_data_method(self, *args, **kargs):
        raise RuntimeError(
            "Data access methods are unsupported. Use normal Python attributes instead"
        )

    # Generic data methods are not needed in python as it can be handled
    # with standard attribute access: https://bugzilla.gnome.org/show_bug.cgi?id=641944
    get_data = _unsupported_data_method
    get_qdata = _unsupported_data_method
    set_data = _unsupported_data_method
    steal_data = _unsupported_data_method
    steal_qdata = _unsupported_data_method
    replace_data = _unsupported_data_method
    replace_qdata = _unsupported_data_method

    # The following methods as unsupported until we verify
    # they work as gi methods.
    bind_property_full = _unsupported_method
    compat_control = _unsupported_method
    interface_find_property = _unsupported_method
    interface_install_property = _unsupported_method
    interface_list_properties = _unsupported_method
    notify_by_pspec = _unsupported_method
    watch_closure = _unsupported_method

    # Make all reference management methods private but still accessible.
    def _ref(self):
        """Deprecated, do not explicitly reference GObjects."""
        warnings.warn(self._ref.__doc__, PyGIDeprecationWarning, stacklevel=2)
        return GObjectModule.Object.ref(self)

    def _ref_sink(self):
        """Deprecated, do not explicitly reference GObjects."""
        warnings.warn(self._ref_sink.__doc__, PyGIDeprecationWarning, stacklevel=2)
        return GObjectModule.Object.ref_sink(self)

    def _unref(self):
        """Deprecated, do not explicitly reference GObjects."""
        warnings.warn(self._unref.__doc__, PyGIDeprecationWarning, stacklevel=2)
        return GObjectModule.Object.unref(self)

    def _force_floating(self):
        """Deprecated, do not explicitly float GObjects."""
        warnings.warn(
            self._force_floating.__doc__, PyGIDeprecationWarning, stacklevel=2
        )
        return GObjectModule.Object.force_floating(self)

    ref = _unsupported_method
    ref_sink = _unsupported_method
    unref = _unsupported_method
    force_floating = _unsupported_method

    # The following methods are static APIs which need to leap frog the
    # gi methods until we verify the gi methods can replace them.
    get_property = _gi.GObject.get_property
    get_properties = _gi.GObject.get_properties
    set_property = _gi.GObject.set_property
    set_properties = _gi.GObject.set_properties
    bind_property = _gi.GObject.bind_property
    connect = _gi.GObject.connect
    connect_after = _gi.GObject.connect_after
    connect_object = _gi.GObject.connect_object
    connect_object_after = _gi.GObject.connect_object_after
    disconnect_by_func = _gi.GObject.disconnect_by_func
    handler_block_by_func = _gi.GObject.handler_block_by_func
    handler_unblock_by_func = _gi.GObject.handler_unblock_by_func
    emit = _gi.GObject.emit
    chain = _gi.GObject.chain
    weak_ref = _gi.GObject.weak_ref
    __copy__ = _gi.GObject.__copy__
    __deepcopy__ = _gi.GObject.__deepcopy__

    def freeze_notify(self) -> typing.ContextManager[None]:
        """Freezes the object's property-changed notification queue.

        :returns:
            A context manager which optionally can be used to
            automatically thaw notifications.

        This will freeze the object so that "notify" signals are blocked until
        the thaw_notify() method is called.

        .. code-block:: python

            with obj.freeze_notify():
                pass
        """
        super().freeze_notify()
        return _FreezeNotifyManager(self)

    def connect_data(
        self,
        detailed_signal: str,
        handler: typing.Callable,
        *data: typing.Any,
        connect_flags: ConnectFlags = 0,
    ) -> int:
        """Connect a callback to the given signal with optional user data.

        :param str detailed_signal:
            A detailed signal to connect to.
        :param callable handler:
            Callback handler to connect to the signal.
        :param *data:
            Variable data which is passed through to the signal handler.
        :param GObject.ConnectFlags connect_flags:
            Flags used for connection options.
        :returns:
            A signal id which can be used with disconnect.
        """
        if connect_flags & GObjectModule.ConnectFlags.AFTER:
            connect_func = _gi.GObject.connect_after
        else:
            connect_func = _gi.GObject.connect

        if connect_flags & GObjectModule.ConnectFlags.SWAPPED:
            if len(data) != 1:
                raise ValueError(
                    "Using GObject.ConnectFlags.SWAPPED requires exactly "
                    "one argument for user data, got: %s" % [data]
                )

            def new_handler(obj, *args):
                # Swap obj with the last element in args which will be the user
                # data passed to the connect function.
                args = list(args)
                swap = args.pop()
                args = [*args, obj]
                return handler(swap, *args)
        else:
            new_handler = handler

        return connect_func(self, detailed_signal, new_handler, *data)

    #
    # Aliases
    #

    handler_block = signal_handler_block
    handler_unblock = _signalmethod(GObjectModule.signal_handler_unblock)
    disconnect = _signalmethod(GObjectModule.signal_handler_disconnect)
    handler_disconnect = _signalmethod(GObjectModule.signal_handler_disconnect)
    handler_is_connected = _signalmethod(GObjectModule.signal_handler_is_connected)
    stop_emission_by_name = _signalmethod(GObjectModule.signal_stop_emission_by_name)

    #
    # Deprecated Methods
    #

    def stop_emission(self, detailed_signal: str) -> None:
        """Deprecated, please use stop_emission_by_name."""
        warnings.warn(self.stop_emission.__doc__, PyGIDeprecationWarning, stacklevel=2)
        self.stop_emission_by_name(detailed_signal)

    emit_stop_by_name = stop_emission


Object = override(Object)
GObject = Object
__all__ += ["GObject", "Object"]


class Binding(GObjectModule.Binding):
    def __call__(self):
        warnings.warn(
            "Using parentheses (binding()) to retrieve the Binding object is no "
            'longer needed because the binding is returned directly from "bind_property.',
            PyGIDeprecationWarning,
            stacklevel=2,
        )
        return self


Binding = override(Binding)
__all__.append("Binding")


Property = propertyhelper.Property
Signal = signalhelper.Signal
SignalOverride = signalhelper.SignalOverride
# Deprecated naming "property" available for backwards compatibility.
# Keep this at the end of the file to avoid clobbering the builtin.
property = Property
deprecated_attr("GObject", "property", "GObject.Property")
__all__ += ["Property", "Signal", "SignalOverride", "property"]


@override
class ParamSpec(GObjectModule.ParamSpec):
    @property
    def nick(self) -> str:
        return self._nick

    @nick.setter
    def nick(self, nick: str) -> None:
        self._nick = nick

    @property
    def blurb(self) -> str:
        return self._blurb

    @blurb.setter
    def blurb(self, blurb: str) -> None:
        self._blurb = blurb


GParamSpec = ParamSpec
deprecated_attr("GObject", "GParamSpec", "GObject.ParamSpec")
__all__ += ["GParamSpec", "ParamSpec"]


@override
class ParamSpecEnum(GObjectModule.ParamSpecEnum):
    @property
    def enum_class(self):
        gtype = super().enum_class.g_type_class.g_type
        if pytype := gtype.pytype:
            return pytype
        return _gi.enum_add(None, gtype.name, gtype, None)


__all__.append("ParamSpecEnum")


@override
class ParamSpecFlags(GObjectModule.ParamSpecFlags):
    @property
    def flags_class(self):
        gtype = super().flags_class.g_type_class.g_type
        if pytype := gtype.pytype:
            return pytype
        return _gi.flags_add(None, gtype.name, gtype, None)


__all__.append("ParamSpecFlags")


class Float(float):
    """A wrapper to force conversion to G_TYPE_FLOAT instead of G_TYPE_DOUBLE when
    used in GValue APIs.
    """

    __gtype__ = TYPE_FLOAT


__all__.append("Float")



# Copyright (C) 2009 Johan Dahlin <johan@gnome.org>
#               2010 Simon van der Linden <svdlinden@src.gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import sys
import warnings

from ..overrides import override, strip_boolean_result
from ..module import get_introspection_module
from gi import PyGIDeprecationWarning, require_version, get_option

Gdk = get_introspection_module("Gdk")
GDK3 = Gdk._version == "3.0"
GDK4 = Gdk._version == "4.0"

__all__ = []


# https://bugzilla.gnome.org/show_bug.cgi?id=673396
try:
    require_version("GdkX11", Gdk._version)
    from gi.repository import GdkX11

    GdkX11
except (ValueError, ImportError):
    pass

if GDK3:
    # Gdk.Color was deprecated since 3.14 and dropped in Gtk-4.0
    class Color(Gdk.Color):
        MAX_VALUE = 65535

        def __init__(self, red, green, blue):
            Gdk.Color.__init__(self)
            self.red = red
            self.green = green
            self.blue = blue

        def __eq__(self, other):
            if not isinstance(other, Gdk.Color):
                return False
            return self.equal(other)

        # This is required (even when __eq__ is defined) in order
        # for != operator to work as expected
        def __ne__(self, other):
            return not self == other

        def __repr__(self):
            return f"Gdk.Color(red={self.red}, green={self.green}, blue={self.blue})"

        red_float = property(
            fget=lambda self: self.red / float(self.MAX_VALUE),
            fset=lambda self, v: setattr(self, "red", int(v * self.MAX_VALUE)),
        )

        green_float = property(
            fget=lambda self: self.green / float(self.MAX_VALUE),
            fset=lambda self, v: setattr(self, "green", int(v * self.MAX_VALUE)),
        )

        blue_float = property(
            fget=lambda self: self.blue / float(self.MAX_VALUE),
            fset=lambda self, v: setattr(self, "blue", int(v * self.MAX_VALUE)),
        )

        def to_floats(self):
            """Return (red_float, green_float, blue_float) triple."""
            return (self.red_float, self.green_float, self.blue_float)

        @staticmethod
        def from_floats(red, green, blue):
            """Return a new Color object from red/green/blue values from 0.0 to 1.0."""
            return Color(
                int(red * Color.MAX_VALUE),
                int(green * Color.MAX_VALUE),
                int(blue * Color.MAX_VALUE),
            )

    Color = override(Color)
    __all__.append("Color")


# Introduced since Gtk-3.0
class RGBA(Gdk.RGBA):
    def __init__(self, red=1.0, green=1.0, blue=1.0, alpha=1.0):
        Gdk.RGBA.__init__(self)
        self.red = red
        self.green = green
        self.blue = blue
        self.alpha = alpha

    def __eq__(self, other):
        if not isinstance(other, Gdk.RGBA):
            return False
        return self.equal(other)

    # This is required (even when __eq__ is defined) in order
    # for != operator to work as expected
    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return f"Gdk.RGBA(red={self.red:f}, green={self.green:f}, blue={self.blue:f}, alpha={self.alpha:f})"

    def __iter__(self):
        """Iterator which allows easy conversion to tuple and list types."""
        yield self.red
        yield self.green
        yield self.blue
        yield self.alpha

    if GDK3:

        def to_color(self):
            """Converts this RGBA into a Color instance which excludes alpha."""
            return Color(
                int(self.red * Color.MAX_VALUE),
                int(self.green * Color.MAX_VALUE),
                int(self.blue * Color.MAX_VALUE),
            )

        @classmethod
        def from_color(cls, color):
            """Returns a new RGBA instance given a Color instance."""
            return cls(color.red_float, color.green_float, color.blue_float)


RGBA = override(RGBA)
__all__.append("RGBA")


if GDK3:
    # Newer GTK/gobject-introspection (3.17.x) include GdkRectangle in the
    # typelib. See https://bugzilla.gnome.org/show_bug.cgi?id=748832 and
    # https://bugzilla.gnome.org/show_bug.cgi?id=748833
    if not hasattr(Gdk, "Rectangle"):
        from gi.repository import cairo as _cairo

        Rectangle = _cairo.RectangleInt

        __all__.append("Rectangle")
    else:
        # https://bugzilla.gnome.org/show_bug.cgi?id=756364
        # These methods used to be functions, keep aliases for backwards compat
        rectangle_intersect = Gdk.Rectangle.intersect
        rectangle_union = Gdk.Rectangle.union

        __all__.append("rectangle_intersect")
        __all__.append("rectangle_union")

if GDK3:

    class Window(Gdk.Window):
        def __new__(cls, parent, attributes, attributes_mask):
            # Gdk.Window had to be made abstract,
            # this override allows using the standard constructor
            return Gdk.Window.new(parent, attributes, attributes_mask)

        def __init__(self, parent, attributes, attributes_mask):
            pass

        def cairo_create(self):
            return Gdk.cairo_create(self)

    Window = override(Window)
    __all__.append("Window")

if GDK3:
    Gdk.EventType._2BUTTON_PRESS = getattr(Gdk.EventType, "2BUTTON_PRESS")
    Gdk.EventType._3BUTTON_PRESS = getattr(Gdk.EventType, "3BUTTON_PRESS")

    class Event(Gdk.Event):
        _UNION_MEMBERS = {
            Gdk.EventType.DELETE: "any",
            Gdk.EventType.DESTROY: "any",
            Gdk.EventType.MOTION_NOTIFY: "motion",
            Gdk.EventType.BUTTON_PRESS: "button",
            Gdk.EventType.BUTTON_RELEASE: "button",
            Gdk.EventType.KEY_PRESS: "key",
            Gdk.EventType.KEY_RELEASE: "key",
            Gdk.EventType.ENTER_NOTIFY: "crossing",
            Gdk.EventType.LEAVE_NOTIFY: "crossing",
            Gdk.EventType.FOCUS_CHANGE: "focus_change",
            Gdk.EventType.CONFIGURE: "configure",
            Gdk.EventType.PROXIMITY_IN: "proximity",
            Gdk.EventType.PROXIMITY_OUT: "proximity",
            Gdk.EventType.DRAG_ENTER: "dnd",
            Gdk.EventType.DRAG_LEAVE: "dnd",
            Gdk.EventType.DRAG_MOTION: "dnd",
            Gdk.EventType.DROP_START: "dnd",
            Gdk.EventType._2BUTTON_PRESS: "button",
            Gdk.EventType._3BUTTON_PRESS: "button",
            Gdk.EventType.PROPERTY_NOTIFY: "property",
            Gdk.EventType.SELECTION_CLEAR: "selection",
            Gdk.EventType.SELECTION_REQUEST: "selection",
            Gdk.EventType.SELECTION_NOTIFY: "selection",
            Gdk.EventType.DRAG_STATUS: "dnd",
            Gdk.EventType.DROP_FINISHED: "dnd",
            Gdk.EventType.CLIENT_EVENT: "client",
            Gdk.EventType.VISIBILITY_NOTIFY: "visibility",
            Gdk.EventType.SCROLL: "scroll",
            Gdk.EventType.EXPOSE: "expose",
            Gdk.EventType.MAP: "any",
            Gdk.EventType.UNMAP: "any",
        }

        if hasattr(Gdk.EventType, "TOUCH_BEGIN"):
            _UNION_MEMBERS.update(
                {
                    Gdk.EventType.TOUCH_BEGIN: "touch",
                    Gdk.EventType.TOUCH_UPDATE: "touch",
                    Gdk.EventType.TOUCH_END: "touch",
                    Gdk.EventType.TOUCH_CANCEL: "touch",
                }
            )

        def __getattr__(self, name):
            real_event = getattr(self, "_UNION_MEMBERS").get(self.type)
            if real_event:
                return getattr(getattr(self, real_event), name)
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            )

        def __setattr__(self, name, value):
            real_event = getattr(self, "_UNION_MEMBERS").get(self.type)
            if real_event:
                setattr(getattr(self, real_event), name, value)
            else:
                Gdk.Event.__setattr__(self, name, value)

        def __repr__(self):
            base_repr = Gdk.Event.__repr__(self).strip("><")
            return f"<{base_repr} type={self.type!r}>"

    Event = override(Event)
    __all__.append("Event")

    # manually bind GdkEvent members to GdkEvent

    modname = globals()["__name__"]
    module = sys.modules[modname]

    # right now we can't get the type_info from the
    # field info so manually list the class names
    event_member_classes = [
        "EventAny",
        "EventExpose",
        "EventMotion",
        "EventButton",
        "EventScroll",
        "EventKey",
        "EventCrossing",
        "EventFocus",
        "EventConfigure",
        "EventProximity",
        "EventDND",
        "EventSetting",
        "EventGrabBroken",
        "EventVisibility",
        "EventProperty",
        "EventSelection",
        "EventOwnerChange",
        "EventWindowState",
        "EventVisibility",
    ]

    if hasattr(Gdk, "EventTouch"):
        event_member_classes.append("EventTouch")

    # whitelist all methods that have a success return we want to mask
    gsuccess_mask_funcs = ["get_state", "get_axis", "get_coords", "get_root_coords"]

    for event_class in event_member_classes:
        override_class = type(event_class, (getattr(Gdk, event_class),), {})
        # add the event methods
        for method_info in Gdk.Event.__info__.get_methods():
            name = method_info.get_name()
            event_method = getattr(Gdk.Event, name)

            # use the _gsuccess_mask decorator if this method is whitelisted
            if name in gsuccess_mask_funcs:
                event_method = strip_boolean_result(event_method)
            setattr(override_class, name, event_method)

        setattr(module, event_class, override_class)
        __all__.append(event_class)

    # end GdkEvent overrides

    class DragContext(Gdk.DragContext):
        def finish(self, success, del_, time):
            Gtk = get_introspection_module("Gtk")
            Gtk.drag_finish(self, success, del_, time)

    DragContext = override(DragContext)
    __all__.append("DragContext")

    class Cursor(Gdk.Cursor):
        def __new__(cls, *args, **kwds):
            arg_len = len(args)
            kwd_len = len(kwds)
            total_len = arg_len + kwd_len

            if total_len == 1:
                # Since g_object_newv (super.__new__) does not seem valid for
                # direct use with GdkCursor, we must assume usage of at least
                # one of the C constructors to be valid.
                return cls.new(*args, **kwds)

            if total_len == 2:
                warnings.warn(
                    'Calling "Gdk.Cursor(display, cursor_type)" has been deprecated. '
                    "Please use Gdk.Cursor.new_for_display(display, cursor_type). "
                    "See: https://pygobject.gnome.org/devguide/initializer_deprecations.html",
                    PyGIDeprecationWarning,
                )
                return cls.new_for_display(*args, **kwds)

            if total_len == 4:
                warnings.warn(
                    'Calling "Gdk.Cursor(display, pixbuf, x, y)" has been deprecated. '
                    "Please use Gdk.Cursor.new_from_pixbuf(display, pixbuf, x, y). "
                    "See: https://pygobject.gnome.org/devguide/initializer_deprecations.html",
                    PyGIDeprecationWarning,
                )
                return cls.new_from_pixbuf(*args, **kwds)

            raise ValueError("Wrong number of parameters")

    Cursor = override(Cursor)
    __all__.append("Cursor")

    # Gdk.Color was deprecated since 3.14 and dropped in Gtk-4.0
    color_parse = strip_boolean_result(Gdk.color_parse)
    __all__.append("color_parse")

    # Note, we cannot override the entire class as Gdk.Atom has no gtype, so just
    # hack some individual methods
    def _gdk_atom_str(atom):
        n = atom.name()
        if n:
            return n
        # fall back to atom index
        return f"Gdk.Atom<{hash(atom):d}>"

    def _gdk_atom_repr(atom):
        n = atom.name()
        if n:
            return f'Gdk.Atom.intern("{n}", False)'
        # fall back to atom index
        return f"<Gdk.Atom({hash(atom):d})>"

    Gdk.Atom.__str__ = _gdk_atom_str
    Gdk.Atom.__repr__ = _gdk_atom_repr


if GDK4:
    from gi.repository import Gio

    class FileList(Gdk.FileList):
        if hasattr(Gdk.FileList, "new_from_list"):

            def __new__(cls, files):
                files_list = []
                if isinstance(files, (tuple, list)):
                    for f in files:
                        if isinstance(f, Gio.File):
                            files_list.append(f)
                        else:
                            raise TypeError(
                                "Constructor requires a list or tuple of Gio.File instances"
                            )
                else:
                    raise TypeError(
                        "Constructor requires a list or tuple of Gio.File instances"
                    )
                return Gdk.FileList.new_from_list(files)

        def __iter__(self):
            return iter(self.get_files())

        def __len__(self):
            return len(self.get_files())

        def __getitem__(self, index):
            return self.get_files()[index]

    FileList = override(FileList)
    __all__.append("FileList")


# constants
if GDK3:
    SELECTION_PRIMARY = Gdk.atom_intern("PRIMARY", True)
    __all__.append("SELECTION_PRIMARY")

    SELECTION_SECONDARY = Gdk.atom_intern("SECONDARY", True)
    __all__.append("SELECTION_SECONDARY")

    SELECTION_CLIPBOARD = Gdk.atom_intern("CLIPBOARD", True)
    __all__.append("SELECTION_CLIPBOARD")

    TARGET_BITMAP = Gdk.atom_intern("BITMAP", True)
    __all__.append("TARGET_BITMAP")

    TARGET_COLORMAP = Gdk.atom_intern("COLORMAP", True)
    __all__.append("TARGET_COLORMAP")

    TARGET_DRAWABLE = Gdk.atom_intern("DRAWABLE", True)
    __all__.append("TARGET_DRAWABLE")

    TARGET_PIXMAP = Gdk.atom_intern("PIXMAP", True)
    __all__.append("TARGET_PIXMAP")

    TARGET_STRING = Gdk.atom_intern("STRING", True)
    __all__.append("TARGET_STRING")

    SELECTION_TYPE_ATOM = Gdk.atom_intern("ATOM", True)
    __all__.append("SELECTION_TYPE_ATOM")

    SELECTION_TYPE_BITMAP = Gdk.atom_intern("BITMAP", True)
    __all__.append("SELECTION_TYPE_BITMAP")

    SELECTION_TYPE_COLORMAP = Gdk.atom_intern("COLORMAP", True)
    __all__.append("SELECTION_TYPE_COLORMAP")

    SELECTION_TYPE_DRAWABLE = Gdk.atom_intern("DRAWABLE", True)
    __all__.append("SELECTION_TYPE_DRAWABLE")

    SELECTION_TYPE_INTEGER = Gdk.atom_intern("INTEGER", True)
    __all__.append("SELECTION_TYPE_INTEGER")

    SELECTION_TYPE_PIXMAP = Gdk.atom_intern("PIXMAP", True)
    __all__.append("SELECTION_TYPE_PIXMAP")

    SELECTION_TYPE_WINDOW = Gdk.atom_intern("WINDOW", True)
    __all__.append("SELECTION_TYPE_WINDOW")

    SELECTION_TYPE_STRING = Gdk.atom_intern("STRING", True)
    __all__.append("SELECTION_TYPE_STRING")


if GDK4 and not hasattr(Gdk.PaintableFlags, "SIZE"):
    Gdk.PaintableFlags.SIZE = getattr(Gdk.PaintableFlags, "STATIC_SIZE")
    Gdk.PaintableFlags.CONTENTS = getattr(Gdk.PaintableFlags, "STATIC_CONTENTS")


if GDK3 and get_option("legacy_autoinit"):
    import sys

    initialized, argv = Gdk.init_check(sys.argv)



# Copyright 2025 Marco Trevisan
# SPDX-License-Identifier: LGPL-2.1-or-later

from ..module import get_introspection_module
from ..overrides import deprecated_attr

GLibUnix = get_introspection_module("GLibUnix")

__all__ = []

if hasattr(GLibUnix, "signal_add"):
    signal_add_full = GLibUnix.signal_add
    __all__.append("signal_add_full")
    deprecated_attr("GLibUnix", "signal_add_full", "GLibUnix.signal_add")



# Copyright 2017 Christoph Reiter
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http://www.gnu.org/licenses/>.


import os
import socket
import signal
import asyncio
import threading
from contextlib import closing, contextmanager

from . import _gi


def ensure_socket_not_inheritable(sock):
    """Ensures that the socket is not inherited by child processes.

    Raises:
        EnvironmentError
        NotImplementedError: With Python <3.4 on Windows

    """
    if hasattr(sock, "set_inheritable"):
        sock.set_inheritable(False)
    else:
        try:
            import fcntl
        except ImportError:
            raise NotImplementedError("Not implemented for older Python on Windows")
        else:
            fd = sock.fileno()
            flags = fcntl.fcntl(fd, fcntl.F_GETFD)
            fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)


_wakeup_fd_is_active = False
"""Since we can't check if set_wakeup_fd() is already used for nested event
loops without introducing a race condition we keep track of it globally.
"""


@contextmanager
def wakeup_on_signal():
    """A decorator for functions which create a glib event loop to keep
    Python signal handlers working while the event loop is idling.

    In case an OS signal is received will wake the default event loop up
    shortly so that any registered Python signal handlers registered through
    signal.signal() can run.

    In case the wrapped function is not called from the main thread it will be
    called as is and it will not wake up the default loop for signals.
    """
    global _wakeup_fd_is_active

    if _wakeup_fd_is_active:
        yield
        return

    from gi.repository import GLib

    read_socket, write_socket = socket.socketpair()
    with closing(read_socket), closing(write_socket):
        for sock in [read_socket, write_socket]:
            sock.setblocking(False)
            ensure_socket_not_inheritable(sock)

        try:
            orig_fd = signal.set_wakeup_fd(write_socket.fileno())
        except ValueError:
            # Raised in case this is not the main thread -> give up.
            yield
            return
        else:
            _wakeup_fd_is_active = True

        def signal_notify(source, condition):
            if condition & GLib.IO_IN:
                try:
                    return bool(read_socket.recv(1))
                except OSError:
                    return False
                return True
            return False

        try:
            if os.name == "nt":
                channel = GLib.IOChannel.win32_new_socket(read_socket.fileno())
            else:
                channel = GLib.IOChannel.unix_new(read_socket.fileno())

            source_id = GLib.io_add_watch(
                channel,
                GLib.PRIORITY_DEFAULT,
                (
                    GLib.IOCondition.IN
                    | GLib.IOCondition.HUP
                    | GLib.IOCondition.NVAL
                    | GLib.IOCondition.ERR
                ),
                signal_notify,
            )
            try:
                yield
            finally:
                GLib.source_remove(source_id)
        finally:
            write_fd = signal.set_wakeup_fd(orig_fd)
            if write_fd != write_socket.fileno():
                # Someone has called set_wakeup_fd while func() was active,
                # so let's re-revert again.
                signal.set_wakeup_fd(write_fd)
            _wakeup_fd_is_active = False


PyOS_getsig = _gi.pyos_getsig
PyOS_setsig = _gi.pyos_setsig

# We save the signal pointer so we can detect if glib has changed the
# signal handler behind Python's back (GLib.unix_signal_add)
if signal.getsignal(signal.SIGINT) is signal.default_int_handler:
    startup_sigint_ptr = PyOS_getsig(signal.SIGINT)
else:
    # Something has set the handler before import, we can't get a ptr
    # for the default handler so make sure the pointer will never match.
    startup_sigint_ptr = -1


def sigint_handler_is_default():
    """Returns if on SIGINT the default Python handler would be called."""
    return (
        signal.getsignal(signal.SIGINT) is signal.default_int_handler
        and PyOS_getsig(signal.SIGINT) == startup_sigint_ptr
    )


@contextmanager
def sigint_handler_set_and_restore_default(handler):
    """Context manager for saving/restoring the SIGINT handler default state.

    Will only restore the default handler again if the handler is not changed
    while the context is active.
    """
    assert sigint_handler_is_default()

    signal.signal(signal.SIGINT, handler)
    sig_ptr = PyOS_getsig(signal.SIGINT)
    try:
        yield
    finally:
        if (
            signal.getsignal(signal.SIGINT) is handler
            and PyOS_getsig(signal.SIGINT) == sig_ptr
        ):
            signal.signal(signal.SIGINT, signal.default_int_handler)


def is_main_thread():
    """Returns True in case the function is called from the main thread."""
    return threading.current_thread().name == "MainThread"


_callback_stack = []
_sigint_called = False


@contextmanager
def register_sigint_fallback(callback):
    """Installs a SIGINT signal handler in case the default Python one is
    active which calls 'callback' in case the signal occurs.

    Only does something if called from the main thread.

    In case of nested context managers the signal handler will be only
    installed once and the callbacks will be called in the reverse order
    of their registration.

    The old signal handler will be restored in case no signal handler is
    registered while the context is active.
    """
    # To handle multiple levels of event loops we need to call the last
    # callback first, wait until the inner most event loop returns control
    # and only then call the next callback, and so on... until we
    # reach the outer most which manages the signal handler and raises
    # in the end

    global _callback_stack, _sigint_called

    if not is_main_thread():
        yield
        return

    if not sigint_handler_is_default():
        if _callback_stack:
            # This is an inner event loop, append our callback
            # to the stack so the parent context can call it.
            _callback_stack.append(callback)
            try:
                yield
            finally:
                cb = _callback_stack.pop()
                if _sigint_called:
                    cb()
        else:
            # There is a signal handler set by the user, just do nothing
            yield
        return

    _sigint_called = False

    def sigint_handler(sig_num, frame):
        global _callback_stack, _sigint_called

        if _sigint_called:
            return
        _sigint_called = True
        _callback_stack.pop()()

    _callback_stack.append(callback)
    try:
        with sigint_handler_set_and_restore_default(sigint_handler):
            yield
    finally:
        if _sigint_called:
            signal.default_int_handler(signal.SIGINT, None)
        else:
            _callback_stack.pop()


class DummyEventLoop:
    @classmethod
    @contextmanager
    def paused(cls):
        yield

    @classmethod
    @contextmanager
    def running(cls, quit_func):
        with wakeup_on_signal():
            yield


def get_event_loop(ctx):
    """Return the correct GLibEventLoop or a dummy that just registers the
    signal wakeup mechanism.
    """
    from . import events

    # Try to use the running loop. If there is none, get the policy and
    # try getting one in the hope that this will give us an event loop for the
    # correct context.
    loop = asyncio._get_running_loop()
    if loop is None:
        # NOTE: We intentionally ignore the event loop policy.
        # It is removed in python 3.16 and we cannot properly check if one was
        # set by the user. Also, it is fair to assume that if one is set, then
        # it will be ours.
        loop = events.GLibEventLoopPolicy._get_event_loop()

    if (
        isinstance(loop, events.GLibEventLoop)
        and ctx is not None
        and hash(loop._context) == hash(ctx)
    ):
        return loop

    return DummyEventLoop



# -*- Mode: Python; py-indent-offset: 4 -*-
# pygobject - Python bindings for the GObject library
# Copyright (C) 2021 Benjamin Berg <bberg@redhat.com
# Copyright (C) 2019 James Henstridge <james@jamesh.id.au>
#
#   gi/asyncio.py: GObject asyncio integration
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http://www.gnu.org/licenses/>.

__all__ = ["GLibEventLoop", "GLibEventLoopPolicy"]

import contextlib
import sys
import asyncio
from asyncio import coroutines
import signal
import threading
import selectors
import weakref
import warnings
from collections.abc import Mapping
from contextlib import contextmanager
from . import _ossighelper

from gi.repository import GLib

try:
    g_main_loop_run = super(GLib.MainLoop, GLib.MainLoop).run
except AttributeError:
    g_main_loop_run = GLib.MainLoop.run


class _IdleSource(GLib.Source):
    """Internal helper source for idle task handling

    The only advantage is that we can keep the source around.
    """

    def __init__(self, loop):
        super().__init__()

        self._loop = loop
        # _may_iterate will be False anyway, but might as well set it
        self.set_can_recurse(False)

    def prepare(self):
        if not self._loop._may_iterate:
            return False, -1

        return bool(self._loop._idle_tasks), -1

    def check(self):
        if not self._loop._may_iterate:
            return False

        return bool(self._loop._idle_tasks)

    def dispatch(self, callback, args):
        self._loop._glib_idle_dispatch()
        return GLib.SOURCE_CONTINUE


class GLibTask(asyncio.Task):
    """This is a simple asyncio.Task subclass that will be returned when using the
    GLibEventLoop. It adds functionality to set the priority that is used to
    iterate the task's coroutine.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._glib_idle_priority = GLib.PRIORITY_DEFAULT

    def set_priority(self, priority):
        """Set the GLib priority used to iterate the task's coroutine"""
        assert isinstance(priority, int)

        self._glib_idle_priority = priority

    def get_priority(self, priority):
        """Get the GLib priority used to iterate the task's coroutine"""
        return self._glib_idle_priority

    @classmethod
    def _factory(cls, loop, coro, **kwargs):
        return GLibTask(coro, loop=loop, **kwargs)


class _GLibEventLoopMixin:
    """Base functionally required for both proactor and selector.

    The proactor/selector is always available through _selector, and we assume
    it has the following extra functionality that we provide:
     * _source: the GSource subclass
     * _dispatching: boolean whether it is dispatching currently
     * attach/detach: add/remove the GSource from the main context

    In principle, we simply override run_forever to call into GLib, with the
    assumption that a GSource is registered which will then call back into
    the python mainloop _run_once handler when needed. This in turn calls
    self._selector.select(), which means we just need to make sure to return
    our already prepared events at that point.

    If no main_context is passed, then the current default main context will
    be used when available. If the thread does not yet have any main context
    set, then a main context will be created.

    When creating an event loop like this, it should be used as a context
    manager. This ensure that the main context is set for the thread, that
    GLib routines will properly iterate the EventLoop and that no second
    EventLoop for the same main context is created by accident.
    """

    def __init__(self, main_context):
        # This allows creating a reasonable GLibEventLoop for the current
        # thread without needing to pass anything.
        if main_context is None:
            main_context = GLib.MainContext.get_thread_default()

            if main_context is None:
                if threading.current_thread() is threading.main_thread():
                    # If there is none, and we are on the main thread,
                    # then use the default context
                    main_context = GLib.MainContext.default()
                else:
                    # Otherwise, create a new context
                    main_context = GLib.MainContext()

        self._context = main_context
        self._main_loop = GLib.MainLoop.new(self._context, False)
        self._quit_funcs = []
        self._idle_tasks = []
        self._may_iterate = False
        self._loop_enter_count = 0
        self._loop_was_set = False
        self._ctx_was_set = False

    @contextmanager
    def paused(self):
        """This context manager ensures the EventLoop is *not* being iterated.

        It purely exists to handle the case where python code iterates the main
        context more gracefully.
        """
        # Nothing to do if we are not running or dispatched by ourselves
        if not self._may_iterate:
            yield
            return

        try:
            self._may_iterate = False
            self._selector.detach()
            yield
        finally:
            self._may_iterate = True
            self._selector.attach()

    @contextmanager
    def running(self, quit_func):
        """This context manager ensures the EventLoop is marked as running
        while other API is iterating its main context.
        The passed quit function is used to stop all recursion levels when
        stop() is called.
        """
        assert self._context.acquire()

        self._quit_funcs.append(quit_func)
        # Nested main context iteration (by using glib API)
        if self.is_running():
            try:
                yield
            finally:
                self._context.release()
                self._quit_funcs.pop()
                # Stop recursively
                if self._stopping:
                    self._quit_funcs[-1]()
            return

        # cpython >= 3.13 has _run_forever_setup (see also _GLibEventLoopRunMixin)
        if hasattr(self, "_run_forever_setup"):
            self._run_forever_setup()
        else:
            # Outermost nesting
            self._check_closed()
            self._set_coroutine_origin_tracking(self._debug)
            self._thread_id = threading.get_ident()

            old_agen_hooks = sys.get_asyncgen_hooks()
            sys.set_asyncgen_hooks(
                firstiter=self._asyncgen_firstiter_hook,
                finalizer=self._asyncgen_finalizer_hook,
            )
            asyncio._set_running_loop(self)

        try:
            assert not self._selector._source._dispatching
            self._may_iterate = True
            self._selector.attach()
            self._idle_source = _IdleSource(self)
            self._idle_source.attach(self._context)
            self._idle_source.set_name("GLibEventLoop._idle_source")
            if self._idle_tasks:
                self._idle_source.set_priority(self._idle_tasks[0][0])
            with self:
                yield
        finally:
            self._may_iterate = False
            self._idle_source.destroy()
            self._idle_source = None
            self._selector.detach()
            self._context.release()
            self._thread_id = None

            if hasattr(self, "_run_forever_setup"):
                self._run_forever_cleanup()
            else:
                asyncio._set_running_loop(None)
                with contextlib.suppress(AttributeError):
                    self._set_coroutine_origin_tracking(False)
                sys.set_asyncgen_hooks(*old_agen_hooks)

                self._stopping = False

            self._quit_funcs.pop()
            assert len(self._quit_funcs) == 0

    def time(self):
        return GLib.get_monotonic_time() / 1000000

    def _get_timeout_ms(self):
        if not self.is_running():
            warnings.warn(
                "GLibEventLoop is iterated without being marked as running. Missing override or invalid use of existing API!",
                RuntimeWarning,
            )
        if self._stopping is True:
            warnings.warn(
                "GLibEventLoop is not stopping properly. Missing override or invalid use of existing API!",
                RuntimeWarning,
            )
        if self._ready:
            return 0

        if self._scheduled:
            # The time is floor'ed here.
            # Python dispatches everything ready within the next _clock_resolution.
            timeout = int((self._scheduled[0]._when - self.time()) * 1000)
            return timeout if timeout >= 0 else 0

        return -1

    def _call_soon(self, callback, args, context):
        try:
            # Try to access the corresponding Task (or whatever) through the
            # self parameter of the bound method.
            # If _glib_idle_priority does not exist or it is not a bound method
            # then we'll just catch the AttributeError exception.
            priority = callback.__self__._glib_idle_priority
        except AttributeError:
            priority = GLib.PRIORITY_DEFAULT

        if priority == GLib.PRIORITY_DEFAULT:
            # Just use the underlying python dispatch.
            return super()._call_soon(callback, args, context)

        handle = asyncio.Handle(callback, args, self, context)
        self._idle_tasks.append((priority, handle))
        self._idle_tasks.sort(key=lambda x: x[0])

        # Update priority
        self._idle_source.set_priority(self._idle_tasks[0][0])

        return handle

    def _glib_dispatch(self):
        assert self._may_iterate

        # The idle source disables itself and we are in the other which will not recurse
        self._may_iterate = False
        self._run_once()
        self._may_iterate = True

    def _glib_idle_dispatch(self):
        assert self._may_iterate

        # Pause so that the main Source is not going to dispatch
        # Note that this is pretty expensive, we could optimize it by detecting
        # it when it happens and only doing the detach/attach dance if needed.
        with self.paused():
            priority = self._idle_source.get_priority()

            ready_handles = []
            while self._idle_tasks and self._idle_tasks[0][0] == priority:
                ready_handles.append(self._idle_tasks.pop(0)[1])

            for handle in ready_handles:
                handle._run()

            # There are (new) tasks available to run, ensure the priority is correct
            if self._idle_tasks:
                self._idle_source.set_priority(self._idle_tasks[0][0])

    def stop(self):
        # Simply quit the mainloop
        self._stopping = True
        if self._quit_funcs:
            self._quit_funcs[-1]()

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} running={self.is_running()} "
            f"closed={self.is_closed()} debug={self.get_debug()} "
            f"ctx=0x{hash(self._context):X} loop=0x{hash(self._main_loop):X}>"
        )

    # The purpose of these mixins is to se the thread local main context,
    # which is useful for threading without an EventLoopPolicy.
    def __enter__(self):
        # Already entered, assume everything is fine
        if self._loop_enter_count > 0:
            self._loop_enter_count += 1
            return

        if not self.is_running() and asyncio._get_running_loop() is not None:
            raise RuntimeError("Thread already has a python EventLoop")

        # Fetch the current MainContext to verify the state
        ctx = GLib.MainContext.get_thread_default()
        if ctx is None and threading.current_thread() is threading.main_thread():
            ctx = GLib.MainContext.default()

        # Only permit a new EventLoop, if any old one is closed already
        if (
            ctx is not None
            and hash(ctx) in GLibEventLoopPolicy._loops
            and not GLibEventLoopPolicy._loops[hash(ctx)].is_closed()
        ):
            if GLibEventLoopPolicy._loops[hash(ctx)] is not self:
                raise RuntimeError(
                    f"A GLibEventLoop is already registered through the policy ({GLibEventLoopPolicy._loops[hash(ctx)]}, self={self})"
                )

            self._loop_was_set = True
        else:
            GLibEventLoopPolicy._loops[hash(self._context)] = self

        if hash(ctx) != hash(self._context):
            GLib.MainContext.push_thread_default(self._context)
            self._ctx_was_set = True

        self._loop_enter_count += 1

    def __exit__(self, exc_type, exc_value, traceback):
        self._loop_enter_count -= 1
        if self._loop_enter_count == 0:
            if not self._loop_was_set:
                del GLibEventLoopPolicy._loops[hash(self._context)]
            if self._ctx_was_set:
                GLib.MainContext.pop_thread_default(self._context)
            self._loop_was_set = False
            self._ctx_was_set = False


class _GLibEventLoopRunMixin:
    # This class exists so we don't need to copy the ProactorEventLoop.run_forever,
    # instead, we change the MRO using a metaclass, so that super() sees this class
    # when called in ProactorEventLoop.run_forever.
    #
    # This class is only needed for cpython < 3.13.

    def run_forever(self):
        # NOTE: self._check_running was only added in 3.8 (with a typo in 3.7)
        if self.is_running():
            raise RuntimeError("This event loop is already running")

        with (
            _ossighelper.register_sigint_fallback(self._main_loop.quit),
            self.running(self._main_loop.quit),
        ):
            g_main_loop_run(self._main_loop)


class _SourceBase(GLib.Source):
    """Common Source functionality for both unix and win32"""

    def __init__(self, selector):
        super().__init__()

        self._dispatching = False

        # It is *not* safe to run the *python* part of the mainloop recursively.
        # This error must be caught further up in the chain, otherwise the
        # mainloop will be blocking without an obvious reason.
        self.set_can_recurse(False)
        self.set_name("python asyncio integration")

        self._selector = weakref.ref(selector)

        self._ready = []

    def _loop(self):
        return self._selector()._loop

    def dispatch(self, callback, args):
        # Now, wag the dog by its tail
        self._dispatching = True
        try:
            self._loop()._glib_dispatch()
        finally:
            self._dispatching = False

        return GLib.SOURCE_CONTINUE

    def _get_ready(self):
        if not self._dispatching:
            raise RuntimeError(
                "gi.asyncio.Selector.select only works while it is dispatching!"
            )

        ready = self._ready
        self._ready = []
        return ready


class _SelectorMixin:
    """A Mixin for common functionality of the Selector and Proactor."""

    def __init__(self, context, loop):
        super().__init__()

        self._context = context
        self._loop = loop
        self._fd_to_key = {}

        self._source = _Source(self)

    def close(self):
        # See _Selector.unregister
        if self._source and hash(self._source):
            self._source.destroy()
            self._source = None
        super().close()

    def select(self, timeout=None):
        return self._source._get_ready()

    def _real_select(self, timeout=None):
        return super().select(timeout)


if sys.platform != "win32":

    class GLibEventLoop(
        _GLibEventLoopMixin, _GLibEventLoopRunMixin, asyncio.SelectorEventLoop
    ):
        """An asyncio event loop that runs the python mainloop inside GLib.

        Based on the asyncio.SelectorEventLoop
        """

        _GLIB_SIGNALS = {
            signal.SIGHUP,
            signal.SIGINT,
            signal.SIGTERM,
            signal.SIGUSR1,
            signal.SIGUSR2,
            signal.SIGWINCH,
        }

        # This is based on the selector event loop, but never actually runs select()
        # in the strict sense.
        # We use the selector to register all FDs with the main context using our
        # own GSource. For python timeouts/idle equivalent, we directly query them
        # from the context by providing the _get_timeout_ms function that the
        # GSource uses. This in turn accesses _ready and _scheduled to calculate
        # the timeout and whether python can dispatch anything non-FD based yet.
        #
        # The Selector select() method simply returns the information we already
        # collected.
        #
        # The rest is done by the mixin which overrides run_forever to simply
        # iterate the main context.
        def __init__(self, main_context=None):
            _GLibEventLoopMixin.__init__(self, main_context)

            # _UnixSelectorEventLoop uses _signal_handlers, we could do the same,
            # with the difference that close() would clean up the handlers for us.
            self.__signal_handlers = {}

            selector = _Selector(self._context, self)
            asyncio.SelectorEventLoop.__init__(self, selector)

            # Used by run_once to not busy loop if the timeout is floor'ed to zero
            self._clock_resolution = 1e-3

            # Use our custom Task subclass
            self._task_factory = GLibTask._factory

        def add_signal_handler(self, sig, callback, *args):
            """Add a handler for UNIX signal"""
            if coroutines.iscoroutine(callback) or coroutines.iscoroutinefunction(
                callback
            ):
                raise TypeError("coroutines cannot be used with add_signal_handler()")
            self._check_closed()

            # Can be useful while testing failures
            # assert sig != signal.SIGALRM

            if sig not in self._GLIB_SIGNALS:
                return super().add_signal_handler(sig, callback, *args)

            # Pure python demands that there is only one signal handler
            source = self.__signal_handlers.get(sig, (None, None, None))[0]
            if source:
                source.destroy()

            # Set up a new source with a higher priority than our main one
            source = GLib.unix_signal_source_new(sig)
            source.set_name(f"asyncio signal watch for {sig}")
            source.set_priority(GLib.PRIORITY_HIGH)
            source.attach(self._context)
            source.set_callback(self._signal_cb, sig)

            self.__signal_handlers[sig] = (source, callback, args)
            del source
            return None

        def remove_signal_handler(self, sig):
            if sig not in self._GLIB_SIGNALS:
                return super().remove_signal_handler(sig)

            try:
                source, _, _ = self.__signal_handlers[sig]
                del self.__signal_handlers[sig]
                # Really unref the underlying GSource so that GLib resets the signal handler
                source.destroy()
                source._clear_boxed()

                # GLib does not restore the original signal handler.
                # Try to restore the python handler for SIGINT, this makes
                # Ctrl+C work after the mainloop has quit.
                if (
                    sig == signal.SIGINT
                    and _ossighelper.PyOS_getsig(signal.SIGINT) == 0
                    and _ossighelper.startup_sigint_ptr > 0
                ):
                    _ossighelper.PyOS_setsig(
                        signal.SIGINT, _ossighelper.startup_sigint_ptr
                    )

                return True
            except KeyError:
                return False

        def _signal_cb(self, sig):
            _source, cb, args = self.__signal_handlers.get(sig)

            # Pass over to python mainloop
            self.call_soon(cb, *args)

        def close(self):
            super().close()
            for s in list(self.__signal_handlers):
                self.remove_signal_handler(s)

    def _fileobj_to_fd(fileobj):
        # Note: SelectorEventloop should only be passing FDs
        if isinstance(fileobj, int):
            return fileobj
        return fileobj.fileno()

    class _Source(_SourceBase):
        def prepare(self):
            timeout = self._loop()._get_timeout_ms()

            # NOTE: Always return False, FDs are queried in check and the timeout
            #       needs to be rechecked anyway.
            return False, timeout

        def check(self):
            ready = []

            for key in self._selector()._fd_to_key.values():
                condition = self.query_unix_fd(key._tag)
                events = 0
                # ERR/HUP/NVAL trigger both read/write (PRI cannot happen)
                if condition & ~GLib.IOCondition.OUT:
                    events |= selectors.EVENT_READ
                if condition & ~GLib.IOCondition.IN:
                    events |= selectors.EVENT_WRITE
                if events:
                    ready.append((key, events))
            self._ready = ready

            timeout = self._loop()._get_timeout_ms()
            if timeout == 0:
                return True

            return bool(ready)

    class _SelectorKey(selectors.SelectorKey):
        # Subclass to attach _tag
        pass

    class _FileObjectMapping(Mapping):
        def __init__(self, fd_dict):
            self.fd_dict = fd_dict

        def __len__(self):
            return len(self.fd_dict)

        def get(self, fileobj, default=None):
            fd = _fileobj_to_fd(fileobj)
            return self.fd_dict.get(fd, default)

        def __getitem__(self, fileobj):
            value = self.get(fileobj)
            if value is None:
                raise KeyError(f"{fileobj!r} is not registered")
            return value

        def __iter__(self):
            return iter(self.fd_dict)

    class _Selector(_SelectorMixin, selectors.BaseSelector):
        """A Selector for gi.events.GLibEventLoop registering python IO with GLib."""

        def __init__(self, context, loop):
            super().__init__(context, loop)
            self._map = _FileObjectMapping(self._fd_to_key)

        def attach(self):
            self._source.attach(self._loop._context)

        def detach(self):
            self._source.destroy()
            self._source = _Source(self)
            # re-register the keys with the new source
            for key in self._fd_to_key.values():
                self._register_key(key)

        def _register_key(self, key):
            condition = GLib.IOCondition(0)
            if key.events & selectors.EVENT_READ:
                condition |= GLib.IOCondition.IN
            if key.events & selectors.EVENT_WRITE:
                condition |= GLib.IOCondition.OUT
            key._tag = self._source.add_unix_fd(key.fd, condition)

        def register(self, fileobj, events, data=None):
            if (not events) or (
                events & ~(selectors.EVENT_READ | selectors.EVENT_WRITE)
            ):
                raise ValueError(f"Invalid events: {events!r}")

            fd = _fileobj_to_fd(fileobj)
            if fd in self._fd_to_key:
                raise KeyError(f"{fileobj!r} (FD {fd}) is already registered")

            key = _SelectorKey(fileobj, fd, events, data)

            self._register_key(key)

            self._fd_to_key[fd] = key
            return key

        def unregister(self, fileobj):
            # NOTE: may be called after __del__ has been called.
            fd = _fileobj_to_fd(fileobj)
            key = self._fd_to_key[fd]

            # As __del__ might have happened, the source may be an empty shell
            # object and calling a function on it will crash us.
            # Catch this by checking that the contained pointer is not NULL.
            if self._source and hash(self._source):
                self._source.remove_unix_fd(key._tag)
            del self._fd_to_key[fd]

            return key

        # We could override modify, but it is only slightly when the "events" change.

        def get_key(self, fileobj):
            return self._map[fileobj]

        def get_map(self):
            """Return a mapping of file objects or file descriptors to
            selector keys.
            """
            return self._map


else:

    class _PushRunMixinBackMeta(type):
        # This metaclass changes the MRO so that when run_forever is called, it
        # first calls asyncio.ProactorEventLoop and then chains into
        # _GLibEventLoopRunMixin.run_forever using super().
        # The alternative would be to copy asyncio.ProactorEventLoop.run_forever
        def mro(cls):
            mro = type.mro(cls)
            idx = mro.index(_GLibEventLoopRunMixin)

            return [*mro[:idx], mro[idx + 1], mro[idx], *mro[idx + 2 :]]

    class GLibEventLoop(
        _GLibEventLoopMixin,
        _GLibEventLoopRunMixin,
        asyncio.ProactorEventLoop,
        metaclass=_PushRunMixinBackMeta,
    ):
        """An asyncio event loop that runs the python mainloop inside GLib.

        Based on the asyncio.WindowsProactorEventLoopPolicy
        """

        # This is based on the Windows ProactorEventLoop
        def __init__(self, main_context=None):
            _GLibEventLoopMixin.__init__(self, main_context)

            proactor = _Proactor(self._context, self)
            # Sets both self._proactor and self._selector to the proactor
            asyncio.ProactorEventLoop.__init__(self, proactor)

            # Used by run_once to not busy loop if the timeout is floor'ed to zero
            self._clock_resolution = 1e-3

            # Use our custom Task subclass
            self._task_factory = GLibTask._factory

    class _Source(_SourceBase):
        def __init__(self, proactor):
            super().__init__(proactor)

            # None denotes it is disabled (and will also not handle timeouts)
            self._poll_fd = None

        def enable(self):
            assert self._poll_fd is None

            self._poll_fd = GLib.PollFD(self._selector()._iocp, GLib.IO_IN)
            self.add_poll(self._poll_fd)

        def disable(self):
            self.remove_poll(self._poll_fd)
            self._poll_fd = None

        def prepare(self):
            # Disabled, do not handle timeouts either
            if self._poll_fd is None:
                return False, -1

            timeout = self._loop()._get_timeout_ms()

            return bool(self._ready), timeout

        def check(self):
            if self._poll_fd is None:
                return False

            if self._poll_fd.revents:
                self._ready.extend(self._selector()._real_select(0))

            if self._ready:
                return True

            return self._loop()._get_timeout_ms() == 0

    class _Proactor(_SelectorMixin, asyncio.IocpProactor):
        """A Proactor for gi.events.GLibEventLoop registering python IO with GLib."""

        def __init__(self, context, loop):
            super().__init__(context, loop)

            # We always use the same Source on windows, it disables itself
            self._source = _Source(self)
            self._source.attach(context)

        def attach(self):
            self._source.enable()

        def detach(self):
            self._source.disable()


# The following are deprecated in 3.13 and will be removed in 3.16,
# keep current code working that uses it to the point that we can.
# NOTE: Convenient filtering was added in python 3.10, just ignore all warnings
with warnings.catch_warnings():
    AbstractEventLoopPolicy = getattr(asyncio, "AbstractEventLoopPolicy", object)

_set_event_loop_policy = getattr(asyncio, "set_event_loop_policy", lambda: None)
_get_event_loop_policy = getattr(asyncio, "get_event_loop_policy", lambda: None)


class GLibEventLoopPolicy(AbstractEventLoopPolicy):
    """An asyncio event loop policy that runs the GLib main loop.

    NOTE: Python 3.16 is removing the concept of the event loop policy.
    FIXME: say what to do in the future

    The policy allows creating a new EventLoop for threads other than the main
    thread. For the main thread, you can use get_event_loop() to retrieve the
    correct mainloop and run it.

    Note that, unlike GLib, python does not support running the EventLoop
    recursively. You should never iterate the GLib.MainContext from within
    the python EventLoop as doing so prevents asyncio events from being
    dispatched.

    As such, do not use API such as GLib.MainLoop.run or Gtk.Dialog.run.
    Instead use the proper asynchronous patterns to prevent entirely blocking
    asyncio.
    """

    _loops = {}
    # COMPAT: child watchers were removed in cpython 3.12
    _child_watcher = None

    def __init__(self):
        self.__orig_policy = None

    def get_event_loop(self):
        """Get the event loop for the current context.

        Returns an event loop object for the thread default GLib.MainContext
        or in case of the main thread for the default GLib.MainContext.

        An exception will be thrown if there is no GLib.MainContext for the
        current thread. In that case, using new_event_loop() will create a new
        main context and main loop which can subsequently attached to the thread
        by calling set_event_loop().

        Returns a new GLibEventLoop or raises an exception.
        """
        return self._get_event_loop(force_implicit=True)

    def get_event_loop_for_context(self, ctx):
        """Get the event loop for a specific context."""
        return self._get_event_loop_for_context(ctx, force_implicit=True)

    @classmethod
    def _get_event_loop(cls, force_implicit=False):
        # Get the thread default main context
        ctx = GLib.MainContext.get_thread_default()
        # If there is none, and we are on the main thread, then use the default context
        if ctx is None and threading.current_thread() is threading.main_thread():
            ctx = GLib.MainContext.default()

        # We do not create a main context implicitly;
        # we create a mainloop for an existing context though
        if ctx is None:
            if not force_implicit:
                return None

            raise RuntimeError(
                f"There is no main context set for thread {threading.current_thread().name!r}."
            )

        return cls._get_event_loop_for_context(ctx, force_implicit=force_implicit)

    @classmethod
    def _get_event_loop_for_context(cls, ctx, force_implicit=False):
        """Get the event loop for a specific context."""
        # Note: We cannot attach it to ctx, as getting the default will always
        #       return a new python wrapper. But, we can use hash() as that returns
        #       the pointer to the C structure.
        try:
            loop = cls._loops[hash(ctx)]
            if not loop.is_closed():
                return loop
        except KeyError:
            pass

        if not force_implicit:
            with warnings.catch_warnings():
                if not isinstance(_get_event_loop_policy(), GLibEventLoopPolicy):
                    return None

        cls._loops[hash(ctx)] = GLibEventLoop(ctx)
        if cls._child_watcher and ctx == GLib.MainContext.default():
            cls._child_watcher.attach_loop(cls._loops[hash(ctx)])
        return cls._loops[hash(ctx)]

    def set_event_loop(self, loop):
        """Set the event loop for the current context (python thread) to loop.

        This is only permitted if the thread has no thread default main context
        with the main thread using the default main context.
        """
        # Only accept glib event loops, otherwise things will just mess up
        assert loop is None or isinstance(loop, GLibEventLoop)

        ctx = ctx_td = GLib.MainContext.get_thread_default()
        if ctx is None and threading.current_thread() is threading.main_thread():
            ctx = GLib.MainContext.default()

        if loop is None:
            # We do permit unsetting the current loop/context
            old = self._loops.pop(hash(ctx), None)
            if old:
                if hash(old._context) != hash(ctx):
                    warnings.warn(
                        "GMainContext was changed unknowingly by asyncio integration!",
                        RuntimeWarning,
                    )
                if ctx_td:
                    GLib.MainContext.pop_thread_default(ctx_td)
        else:
            # Only allow attaching if the thread has no main context yet
            if ctx:
                raise RuntimeError(
                    f"Thread {threading.current_thread().name!r} already has a main context, "
                    "get_event_loop() will create a new loop if needed"
                )

            GLib.MainContext.push_thread_default(loop._context)
            self._loops[hash(loop._context)] = loop

    def new_event_loop(self):
        """Create and return a new event loop that iterates a new
        GLib.MainContext.
        """
        return GLibEventLoop(GLib.MainContext())

    def __enter__(self):
        with warnings.catch_warnings():
            self.__orig_policy = _get_event_loop_policy()
            _set_event_loop_policy(self)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # We shouldn't have any running loops at this point, and the ones that
        # got created should be closed eventually.
        # Explicitly close all loops here, it is not reasonable for them to be
        # used after we unregister the EventLoopPolicy below.
        for loop in self._loops.values():
            loop.close()

        with warnings.catch_warnings():
            _set_event_loop_policy(self.__orig_policy)

        # Do not supress any exceptions
        return False

    # NOTE: We do *not* provide a GLib based ChildWatcher implementation!
    # This is *intentional* and *required*. The issue is that python provides
    # API which uses wait4() internally. GLib at the same time uses a thread to
    # handle SIGCHLD signals, which causes a race condition resulting in a
    # critical warning.
    # We just provide a reasonable sane child watcher and disallow the user
    # from choosing one as e.g. MultiLoopChildWatcher is problematic.
    #
    # COMPAT: child watchers were removed in python 3.12
    if sys.platform != "win32" and hasattr(
        AbstractEventLoopPolicy, "get_child_watcher"
    ):

        @classmethod
        def get_child_watcher(cls):
            if cls._child_watcher is None:
                cls._child_watcher = asyncio.ThreadedChildWatcher()

                if threading.current_thread() is threading.main_thread():
                    cls._child_watcher.attach_loop(cls._get_event_loop())

            return cls._child_watcher



# Copyright (C) 2010 Paolo Borelli <pborelli@gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

from __future__ import annotations

import typing

from ..overrides import override
from ..module import get_introspection_module

# Typing relies on https://github.com/pygobject/pygobject-stubs.
if typing.TYPE_CHECKING:
    # Import stubs for type checking this file.
    # FIXME: For IDE typing support, you need to copy
    # pygobject-stubs/src/gi-stubs/repository/*.pyi to gi/repository/
    from gi.repository import Pango
    from typing_extensions import Self

    # Type annotations cannot have `Pango.` prefix because they are copied into
    # Pango stubs module which cannot refer to itself. Use type aliases.
    Context = Pango.Context
else:
    from gi.module import get_introspection_module

    Pango = get_introspection_module("Pango")

__all__ = []


class FontDescription(Pango.FontDescription):
    @staticmethod
    def __new__(cls: type[Self], string: str | None = None) -> Self:
        if string is not None:
            return Pango.font_description_from_string(string)  # type: ignore[return-value]
        return Pango.FontDescription.new()  # type: ignore[return-value]


FontDescription = override(FontDescription)
__all__.append("FontDescription")


class Layout(Pango.Layout):
    def __new__(cls: type[Self], context: Context) -> Self:
        return Pango.Layout.new(context)  # type: ignore[return-value]

    def set_markup(self, text: str, length: int = -1) -> None:
        super().set_markup(text, length)

    def set_text(self, text: str, length: int = -1) -> None:
        super().set_text(text, length)


Layout = override(Layout)
__all__.append("Layout")
